(function() {
	var Grit, EventEmitter;

	/*
	 * require in NodeJS's native EventEmitter class,
	 * or emulate it in the browser.
	 */

	if(typeof exports !== 'undefined') {
		EventEmitter = require('events').EventEmitter;
	} else {
		EventEmitter = function() {
			var handlers = {};

			this.on = this.addListener = function(e, callback) {
				if(!handlers.hasOwnProperty(e)) {
					handlers[e] = [];
				}
				handlers[e].push(callback);
			};

			this.removeListener = function(e, callback) {
				var handleList, index, length, curr;
				if(handlers.hasOwnProperty(e)) {
					handleList = handlers[e];
					for(index = 0, length = handleList.length; index < length; index++) {
						curr = handleList[index];
						if(curr === callback) {
							handleList.splice(index, 1);
							if(handleList.length === 0)
								delete handlers[e];
							return;
						}	
					}
				}
			};

			this.emit = function(e) {
				var handleList, index, length, args;
				if(handlers.hasOwnProperty(e)) {
					handleList = handlers[e];
					args = Array.prototype.slice(this, arguments).slice(1);
					for(index = 0, length = handleList.length; index < length; index++)
						handleList[index].call(this, args);
				}
			};
		};
	}

	/*
	 * The model class itself.
	 */

	Grit = function(properties) {
		EventEmitter.call(this);
		var data, prop, valid, changeData, 
				validateData, ensurePresenceOfData;

		data = {};

		ensurePresenceOfData = function() {
			var prop;
			if(!this.constructor.hasOwnProperty('required') || !this.constructor.required)
				return true;

			for(prop in this.constructor.required) {
				if(this.constructor.required.hasOwnProperty(prop) && !data.hasOwnProperty(prop))
					return false;
			}
			return true;
		};
		
		validateData = function(that, name, value) {
			var validators, valid, index, length;
			if(!that.constructor.hasOwnProperty('validations') || !that.constructor.validations)
				return true;
			if(!that.constructor.validations.hasOwnProperty(name) || !that.constructor.validations[name])
				return true;
			else {
				validators = that.constructor.validations[name];
				for(index = 0, length = validators.length; index < length; index++) {
					if(!validators[index](value))
						return false;
				}
			}
			return true;
		};

		changeData = function(name, value, silent) {
			if(typeof silent === 'undefined' || !silent)
				this.emit('change:' + name, this, value);
			data[name] = value;
			this.sync(this, data, name);
		};

		this.get = function(name) {
			if(data.hasOwnProperty(name))
				return data[name];
			return null;
		};
		this.set = function(name, value, silent) {
			var valid = validateData(this, name, value) && ensurePresenceOfData.call(this);

			if(valid) {
				changeData.call(this, name, value, silent);
				return value;
			}

			return null;
		};

		if(typeof properties !== 'undefined') {
			valid = true;
			if(this.constructor.hasOwnProperty('required')) {
				for(prop in this.constructor.required) {
					if(!properties.hasOwnProperty(prop) || !validateData(this, prop, properties[prop])) {
						valid = false;
						break;
					}
				}
			}	

			if(valid)
				for(prop in properties)
					if(properties.hasOwnProperty(prop))
						changeData.call(this, prop, properties[prop], false);
		}
	};
	Grit.prototype = new EventEmitter();
	Grit.prototype.constructor = Grit;

	// sync is left empty -- redefine it to automate
	// changes being synced to the server or db.
	Grit.prototype.sync = function(model, data, property) {
	};

	// load is left empty -- redefine it to automate
	// loading things from the server or db. load
	// is never auto-called, unlike sync()
	Grit.prototype.load = function() {
	};

	Grit.extend = function(structFn) {
		structFn.prototype = new Grit();
		structFn.prototype.constructor = structFn;
		for(var prop in Grit) {
			if(Grit.hasOwnProperty(prop))
				structFn[prop] = Grit[prop];
		}
		return structFn;
	};

	Grit.require = function() {
		var args, index; 
		args = Array.prototype.slice.call(arguments, 0);

		if(!this.hasOwnProperty('required'))
			this.required = {};

		for(index = 0; index < args.length; index++)
			this.required[args[index]] = true;
	};
	Grit.unrequire = function() {
		var args, index;
		args = Array.prototype.slice.call(arguments, 0);

		if(!this.hasOwnProperty('required'))
			return;

		for(index = 0; index < args.length; index++) {
			if(this.required.hasOwnProperty(args[index]))
				delete this.required[args[index]];
		}
	};

	Grit.validate = function(name, validator) {
		if(!this.hasOwnProperty('validations'))
			this.validations = {};

		if(!this.validations.hasOwnProperty(name))
			this.validations[name] = [];
		this.validations[name].push(validator);
	};
	Grit.removeValidation = function(name, validator) {
		var validators, index, length;
		if(!this.hasOwnProperty('validations'))
			return;

		if(this.validations.hasOwnProperty(name)) {
			validators = this.validations[name];
			for(index = 0, length = validators.length; index < length; index++) {
				if(validators[index] === validator) {
					validators.splice(index, 1);
					break;
				}
			}

			if(validators.length === 0)
				delete this.validations[name];
		}
	};

	/* 
	 * Export
	 */

	if(typeof exports !== 'undefined') {
		exports = module.exports = Grit;
	} else {
		window.Grit = Grit;
	}
}());
